//
// Testing DMA vs cache behaviour.
//
//! \file main.cpp
//! \author J Vijn
//! \date 20090324 - 20090526
//
// === NOTES ===


#include <stdio.h>
#include <string.h>

#include <nds.h>
#include "all_gfx.h"



#ifndef REG_DMA0SAD
#define REG_DMA0SAD			(*(vu32*)0x040000B0)	//!< DMA 0 Source address
#define REG_DMA0DAD			(*(vu32*)0x040000B4)	//!< DMA 0 Destination address
#define REG_DMA0CNT			(*(vu32*)0x040000B8)	//!< DMA 0 Control

#define REG_DMA1SAD			(*(vu32*)0x040000BC)	//!< DMA 1 Source address
#define REG_DMA1DAD			(*(vu32*)0x040000C0)	//!< DMA 1 Destination address
#define REG_DMA1CNT			(*(vu32*)0x040000C4)	//!< DMA 1 Control

#define REG_DMA2SAD			(*(vu32*)0x040000C8)	//!< DMA 2 Source address
#define REG_DMA2DAD			(*(vu32*)0x040000CC)	//!< DMA 2 Destination address
#define REG_DMA2CNT			(*(vu32*)0x040000D0)	//!< DMA 2 Control

#define REG_DMA3SAD			(*(vu32*)0x040000D4)	//!< DMA 3 Source address
#define REG_DMA3DAD			(*(vu32*)0x040000D8)	//!< DMA 3 Destination address
#define REG_DMA3CNT			(*(vu32*)0x040000DC)	//!< DMA 3 Control
#endif

void keyWait()
{
	while(1)
	{
		swiWaitForVBlank();
		scanKeys();
		if(keysDown())
			break;
	}
}

//! Copy data from \a src to \a dst via DMA in a cache/section safe manner.
/*!	The ARM9's DMA doesn't play well with the cache and can't access 
	ITCM or DTCM. This means that a basic DMA copy may not work as 
	expected. This function flushes or invalidates cache if necessary and 
	will only copy if the ranges are accessible.
	\param src	Source pointer.
	\param dst	Destination pointer.
	\param size	Size (in bytes) to copy.
	\return		True if the copy succeeded.
	\note		It's possible I missed some invalid cases, YHBW.
*/
bool dmaCopySafe(const void *src, void *dst, u32 size)
{
	u32 srca= (u32)src, dsta= (u32)dst;

	// Check TCMs and BIOS (0x01000000, 0x0B000000, 0xFFFF0000).
	//# NOTE: probably incomplete checks.
	if((srca>>24)==0x01 || (srca>>24)==0x0B || (srca>>24)==0xFF)
		return false;
	if((dsta>>24)==0x01 || (dsta>>24)==0x0B || (dsta>>24)==0xFF)
		return false;

	if((srca|dsta) & 1)					// Fail on byte copy.
		return false;
	
	while(REG_DMA3CNT & DMA_BUSY) ;

	if((srca>>24)==0x02)				// Write cache back to memory.
		DC_FlushRange(src, size);

	if((srca|dsta|size) & 3)
		dmaCopyHalfWords(3, src, dst, size);
	else
		dmaCopyWords(3, src, dst, size);

	if((dsta>>24)==0x02)				// Set cache of dst range to 'dirty'
		DC_InvalidateRange(dst, size);

	return true;
}

//! Fill \a dst with \a fill via DMA in a cache/section safe manner.
/*!	The ARM9's DMA doesn't play well with the cache and can't access 
	ITCM or DTCM. This means that a basic DMA fill may not work as 
	expected. This function flushes or invalidates cache if necessary and 
	will only fill if the ranges are accessible.
	\param fill	Fill value.
	\param dst	Destination pointer.
	\param size	Size (in bytes) to copy.
	\return		True if the fill succeeded.
	\note		It's possible I missed some invalid cases, YHBW.
*/
bool dmaFillSafe(u32 fill, void *dst, u32 size)
{
	u32 dsta= (u32)dst;

	// Check TCMs and BIOS (0x01000000, 0x0B000000, 0xFFFF0000).
	//# NOTE: probably incomplete checks.
	if((dsta>>24)==0x01 || (dsta>>24)==0x0B || (dsta>>24)==0xFF)
		return false;

	if(dsta & 1)						// Fail on byte fill.
		return false;
	
	while(REG_DMA3CNT & DMA_BUSY) ;

	if((dsta|size) & 3)
		dmaFillHalfWords(fill, dst, size);
	else
		dmaFillWords(fill, dst, size);

	if((dsta>>24)==0x02)				// Set cache of dst range to 'dirty'
		DC_InvalidateRange(dst, size);

	return true;
}


// --------------------------------------------------------------------
// TESTING ROUTINES
// --------------------------------------------------------------------

/*
	As far as I know, this is how it works.
	The NDS has associative cache, 4kB for data and 8kB for code.
	Cache lines are 32 bytes in size. On a load, a new cache 
	line is initialized. At that point, all loads/stores go there 
	until the cache line is released.
	DMA can't read cache, so if a part of RAM is cached, DMA 
	interaction with that part *may* be problematic.
	To solve this, you need to flush or invalidate the memory range.
	- Flush: write the cache back to memory. See DC_FlushRange()
	- Invalidate: indicate a write to memory, signalling that cache 
	  needs to be updated (?). See DC_InvalidateRange().

*/

struct MiniBmp
{
	u16 width, height;
	u16 *data;
};


static const int X_CPU= 48, Y_CPU= 32, RIGHT_CPU= 48+16, BOTTOM_CPU= 32+16;
static const int X_DMA= 80, Y_DMA= 32, RIGHT_DMA= 80+16, BOTTOM_DMA= 32+16;
static const int SIZE= 16*16*2;


// Double buffer in RAM.
u16 gfxBuffer[16*16] ALIGN(4);

MiniBmp bmpA=	{	16, 16, (u16*)abBitmap				};
MiniBmp bmpB=	{	16, 16, &((u16*)abBitmap)[16*16]	};
MiniBmp bmpBuf= {	16, 16, gfxBuffer					};



// Blit to frame via CPU.
void cpuBlit(const MiniBmp *src, int x, int y)
{
	int ix, iy;
	int srcW= src->width, srcH= src->height;
	u16 *srcD= src->data, *dstD= &VRAM_A[y*256+x];
	for(iy=0; iy<srcH; iy++)
		for(ix=0; ix<srcW; ix++)
			dstD[iy*256+ix]= srcD[iy*srcW+ix];
}

// Blit to frame via DMA.
void dmaBlit(const MiniBmp *src, int x, int y)
{
	int iy;
	int srcW= src->width, srcH= src->height;
	u16 *srcD= src->data, *dstD= &VRAM_A[y*256+x];

	for(iy=0; iy<srcH; iy++)
		dmaCopy(&srcD[iy*srcW], &dstD[iy*256], srcW*2);
}

// Blit to frame via safe DMA.
void dmaBlitSafe(const MiniBmp *src, int x, int y)
{
	int iy;
	int srcW= src->width, srcH= src->height;
	u16 *srcD= src->data, *dstD= &VRAM_A[y*256+x];

	for(iy=0; iy<srcH; iy++)
		dmaCopySafe(&srcD[iy*srcW], &dstD[iy*256], srcW*2);
}

void rect(int left, int top, int right, int bottom, u16 color)
{
	int ix, iy;
	for(iy=top; iy<bottom; iy++)
		for(ix=left; ix<right; ix++)
			VRAM_A[iy*256+ix]= color;
}

void frame(int left, int top, int right, int bottom, u16 color)
{
	int ix, iy;
	for(iy=top; iy<bottom; iy++)
	{
		VRAM_A[iy*256+left   ]= color;
		VRAM_A[iy*256+right-1]= color;
	}
	for(ix=left; ix<right; ix++)
	{
		VRAM_A[top       *256+ix]= color;
		VRAM_A[(bottom-1)*256+ix]= color;
	}
}


void caseInit(const char *str)
{
	consoleClear();
	printf(str);
	printf("\n(key to continue)");

	rect(X_CPU, Y_CPU, RIGHT_CPU, BOTTOM_CPU, RGB5(16, 16, 16));
	frame(X_CPU-2, Y_CPU-2, RIGHT_CPU+2, BOTTOM_CPU+2, RGB5(31, 16, 0));

	rect(X_DMA, Y_DMA, RIGHT_DMA, BOTTOM_DMA, RGB5(16, 16, 16));
	frame(X_DMA-2, Y_DMA-2, RIGHT_DMA+2, BOTTOM_DMA+2, RGB5(31, 16, 0));

	memset(bmpBuf.data, 0, SIZE);
}

void test_cache()
{
	consoleDemoInit();
	videoSetMode(MODE_FB0);
	vramSetBankA(VRAM_A_LCD);

	swiWaitForVBlank();
	scanKeys();
	swiWaitForVBlank();

	DC_FlushAll();
	DC_InvalidateAll();


	// --- Blit directly ---
	//# Works. Duh.
	caseInit("(1)\nStraight blit\nPrediction: ok\n");

	cpuBlit(&bmpA, X_CPU, Y_CPU);
	dmaBlit(&bmpA, X_DMA, Y_DMA);
	
	keyWait();


	// --- Copy B to buffer, blit buffer ---
	//# Prediction: Should still be OK. bmpBuf isn't read from yet, 
	//# so not in cache.
	//# Result:
	//# Seems fine.
	caseInit("(2)\ncpu 'B' -> bmpBuf\nbmpBuf -> vram\nPrediction: ok\n");

	memcpy(bmpBuf.data, bmpB.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);


	keyWait();
	// --- Copy A to buffer (holding B), blit buffer ---
	//# Prediction: bmpBuf is still in cache, so 'A' goes there too, 
	//# making the DMA fail.
	//# Result: cpu ok, DMA fail
	caseInit("(3)\ncpu 'A' -> bmpBuf\nbmpBuf-> vram\nPrediction:DMA fail\n");

	memcpy(bmpBuf.data, bmpA.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Copy B to buffer, flush, blit buffer ---
	//# Prediction: Should be fine.
	//# Result: Both fine.
	caseInit("(4)\ncpu 'B' -> bmpBuf\nflush\nbmpBuf -> vram\nPrediction: ok\n");

	memcpy(bmpBuf.data, bmpB.data, SIZE);

	DC_FlushRange(bmpBuf.data, SIZE);
	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Copy A to buffer, Invalidate, blit buffer ---
	//# Prediction: Should be fine.
	//# Result: BOTH bad in same way. Why? Because the 
	//# writes go to cache, not RAM. The invalidate will update 
	//# cache FROM ram, which actually held the wrong data.
	caseInit("(5)\ncpu 'A' -> bmpBuf\nInvalidate\nbmpBuf -> vram\n"
		"Prediction: fail (corrupt bmpBuf)\n");

	memcpy(bmpBuf.data, bmpA.data, SIZE);

	DC_InvalidateRange(bmpBuf.data, SIZE);
	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Copy A via DMA, blit ---
	//# Prediction: Should require invalidate. cpuBlit to fail?
	//# Result:
	//# cpu fail, DMA ok.
	caseInit("(6)\ndma 'A' -> bmpBuf\nflush\nbmpBuf -> vram\nPrediction: cpu fail\n");

	dmaCopy(bmpA.data, bmpBuf.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- DMA B to buffer, invalidate, blit buffer ---
	//# Prediction: Should be fine.
	//# Result: both ok.
	caseInit("(7)\nDMA 'B' -> bmpBuf\nInvalidate\nbmpBuf -> vram\nPrediction: ok\n");

	dmaCopy(bmpB.data, bmpBuf.data, SIZE);
	DC_InvalidateRange(bmpBuf.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Copy A to buffer, Invalidate/flush, blit buffer ---
	//# Prediction: ???
	//# Result: BOTH bad in same way. Why? Because the 
	//# writes go to cache, not RAM. The invalidate will update 
	//# cache FROM ram, which actually held the wrong data.
	caseInit("(8)\ncpu 'A' -> bmpBuf\nInvalidate+Flush\nbmpBuf -> vram\nPrediction: fail\n");

	memcpy(bmpBuf.data, bmpA.data, SIZE);

	DC_InvalidateRange(bmpBuf.data, SIZE);
	DC_FlushRange(bmpBuf.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlit(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Copy B to buffer, safe-DMA to VRAM ---
	//# Prediction: Should be fine
	//# Result: ok
	caseInit("(9)\ncpu 'B' -> bmpBuf\ndmaSafe bmpBuf -> vram\nPrediction: ok\n");

	memcpy(bmpBuf.data, bmpB.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlitSafe(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Safe-DMA A to buffer, safe-DMA to VRAM ---
	//# Prediction: Should be fine
	//# Result: ok
	caseInit("(10)\ndmaSafe 'A' -> bmpBuf\nbmpBuf -> vram\nPrediction: ok\n");

	dmaCopySafe(bmpA.data, bmpBuf.data, SIZE);

	cpuBlit(&bmpBuf, X_CPU, Y_CPU);
	dmaBlitSafe(&bmpBuf, X_DMA, Y_DMA);

	keyWait();


	// --- Copy A to LOCAL buffer, then blit ---
	//# Prediction: dma fail.
	//# Result: ok
	caseInit("(11)\n'B' -> local buffer\n -> vram\nPrediction: dma fail\n");

	u16 localBuf[16*16];
	MiniBmp bmpLocal= { 16, 16, localBuf };

	memcpy(localBuf, bmpB.data, SIZE);

	cpuBlit(&bmpLocal, X_CPU, Y_CPU);
	dmaBlit(&bmpLocal, X_DMA, Y_DMA);

	keyWait();
}

int main(void)
{
	irqEnable(IRQ_VBLANK);


	while(1)
	{
		test_cache();
		keyWait();
		swiWaitForVBlank();
	}

	return 0;
}
